@@ -18,6 +18,7 @@
  */
 
 #include <gtk/gtk.h>
+#include <gdk/gdkkeysyms.h>
 #include "gailcombobox.h"
 
 #if GTK_MINOR_VERSION > 4
@@ -46,6 +47,8 @@ static gint         gail_combo_box_get_n_actions           (AtkAction      *acti
 ;
 static G_CONST_RETURN gchar* gail_combo_box_get_description(AtkAction      *action,
                                                             gint           i);
+static G_CONST_RETURN gchar* gail_combo_box_get_keybinding   (AtkAction       *action,
+		                                             gint            i);
 static G_CONST_RETURN gchar* gail_combo_box_action_get_name(AtkAction      *action,
                                                             gint           i);
 static gboolean              gail_combo_box_set_description(AtkAction      *action,
@@ -137,6 +140,7 @@ static void
 gail_combo_box_object_init (GailComboBox      *combo_box)
 {
   combo_box->press_description = NULL;
+  combo_box->press_keybinding = NULL;
   combo_box->old_selection = -1;
   combo_box->name = NULL;
   combo_box->popup_set = FALSE;
@@ -241,7 +245,6 @@ gail_combo_box_get_name (AtkObject *obj)
 
   combo_box = GTK_COMBO_BOX (widget);
   gail_combo_box = GAIL_COMBO_BOX (obj);
-  g_free (gail_combo_box->name);
   if (gtk_combo_box_get_active_iter (combo_box, &iter))
     {
       model = gtk_combo_box_get_model (combo_box);
@@ -253,9 +256,14 @@ gail_combo_box_get_name (AtkObject *obj)
           gtk_tree_model_get_value (model, &iter, i, &value);
           if (G_VALUE_HOLDS_STRING (&value))
             {
-              gail_combo_box->name =  (gchar *) g_value_get_string (&value);
+	      if (gail_combo_box->name) g_free (gail_combo_box->name);
+              gail_combo_box->name =  g_strdup ((gchar *) 
+						g_value_get_string (&value));
+	      g_value_unset (&value);
               break;
             }
+	  else
+	    g_value_unset (&value);
         }
     }
   return gail_combo_box->name;
@@ -343,6 +351,7 @@ atk_action_interface_init (AtkActionIface *iface)
   iface->do_action = gail_combo_box_do_action;
   iface->get_n_actions = gail_combo_box_get_n_actions;
   iface->get_description = gail_combo_box_get_description;
+  iface->get_keybinding = gail_combo_box_get_keybinding;
   iface->get_name = gail_combo_box_action_get_name;
   iface->set_description = gail_combo_box_set_description;
 }
@@ -386,17 +395,17 @@ idle_do_action (gpointer data)
   AtkObject *popup;
   gboolean do_popup;
 
+  GDK_THREADS_ENTER ();
+
   gail_combo_box = GAIL_COMBO_BOX (data);
   gail_combo_box->action_idle_handler = 0;
   widget = GTK_ACCESSIBLE (gail_combo_box)->widget;
-  if (widget == NULL)
-    /*
-     * State is defunct
-     */
-    return FALSE;
-
-  if (!GTK_WIDGET_SENSITIVE (widget) || !GTK_WIDGET_VISIBLE (widget))
-    return FALSE;
+  if (widget == NULL || /* State is defunct */
+      !GTK_WIDGET_SENSITIVE (widget) || !GTK_WIDGET_VISIBLE (widget))
+    {
+      GDK_THREADS_LEAVE ();
+      return FALSE;
+    }
 
   combo_box = GTK_COMBO_BOX (widget);
 
@@ -407,6 +416,8 @@ idle_do_action (gpointer data)
   else
       gtk_combo_box_popdown (combo_box);
 
+  GDK_THREADS_LEAVE ();
+
   return FALSE;
 }
 
@@ -434,6 +445,60 @@ gail_combo_box_get_description (AtkAction *action,
     return NULL;
 }
 
+static G_CONST_RETURN gchar*
+gail_combo_box_get_keybinding (AtkAction *action,
+		                    gint      i)
+{
+  GailComboBox *combo_box;
+  gchar *return_value = NULL;
+  combo_box = GAIL_COMBO_BOX (action);
+  switch (i)
+  {
+     case 0:
+      {
+	  GtkWidget *widget;							    
+	  GtkWidget *label;
+	  AtkRelationSet *set;
+	  AtkRelation *relation;							     
+	  GPtrArray *target;
+	  gpointer target_object;
+	  guint key_val;
+
+	  combo_box = GAIL_COMBO_BOX (action);
+	  widget = GTK_ACCESSIBLE (combo_box)->widget;
+	  if (widget == NULL)
+             return NULL;
+	  set = atk_object_ref_relation_set (ATK_OBJECT (action));							
+	  if (!set)
+             return NULL;
+	  label = NULL;
+	  relation = atk_relation_set_get_relation_by_type (set, ATK_RELATION_LABELLED_BY);
+	  if (relation)
+	  {
+	     target = atk_relation_get_target (relation);
+	     target_object = g_ptr_array_index (target, 0);
+	     if (GTK_IS_ACCESSIBLE (target_object))
+	     {
+	        label = GTK_ACCESSIBLE (target_object)->widget;
+	     }
+	  }  
+	  g_object_unref (set);
+	  if (GTK_IS_LABEL (label))
+	  {
+             key_val = gtk_label_get_mnemonic_keyval (GTK_LABEL (label));
+	     if (key_val != GDK_VoidSymbol)
+             return_value = gtk_accelerator_name (key_val, GDK_MOD1_MASK);
+	  }
+	   g_free (combo_box->press_keybinding);
+	   combo_box->press_keybinding = return_value;
+	   break;                                                                                    }                             
+    default:
+	   break;
+  }
+  return return_value;
+}
+
+
 static G_CONST_RETURN gchar*
 gail_combo_box_action_get_name (AtkAction *action,
                                 gint      i)
@@ -610,6 +675,7 @@ gail_combo_box_finalize (GObject *object)
   GailComboBox *combo_box = GAIL_COMBO_BOX (object);
 
   g_free (combo_box->press_description);
+  g_free (combo_box->press_keybinding);
   g_free (combo_box->name);
   if (combo_box->action_idle_handler)
     {
